@@ -284,7 +284,7 @@ static resource resource_constants[] =
 #define RECOVER_END()         ts->recover = __or; \
 			    }
 #define RECOVER(mode, code) \
-        RECOVER_BEGIN(mode); \
+	RECOVER_BEGIN(mode); \
 	code; \
 	RECOVER_END();
 
@@ -366,9 +366,8 @@ discardBuf(string_buffer *b)
     free(b->buf);
 }
 
-
 static inline int
-addBuf(string_buffer *b, int c)
+addBuf_wchar(string_buffer *b, int c)
 { if ( b->in < b->end )
   { *b->in++ = c;
     return TRUE;
@@ -377,6 +376,26 @@ addBuf(string_buffer *b, int c)
   return growBuffer(b, c);
 }
 
+static inline void
+utf16_encode(int c, int *lp, int *tp)
+{ c -= 0x10000;
+  *lp = (c>>10)+0xD800;
+  *tp = (c&0X3FF)+0xDC00;
+}
+
+static inline int
+addBuf(string_buffer *b, int c)
+{
+#if SIZEOF_WCHAR_T == 2
+  if ( c > 0xffff )
+  { int l, t ;
+    utf16_encode(c, &l, &t);
+    return ( addBuf_wchar(b, l) &&
+	     addBuf_wchar(b, t) );
+  } else
+#endif
+  return addBuf_wchar(b, c);
+}
 
 static inline int
 bufSize(string_buffer *b)
@@ -688,7 +707,7 @@ syntax_message(turtle_state *ts, const char *msg, int is_error)
 
   if ( !(ex1=PL_new_term_ref()) ||
        !PL_unify_term(ex1, PL_FUNCTOR, FUNCTOR_syntax_error1,
-		              PL_CHARS, msg) )
+			      PL_CHARS, msg) )
     return FALSE;
 
   return print_message(ts, ex1, is_error);
@@ -1089,7 +1108,7 @@ make_absolute_resource(turtle_state *ts, const wchar_t *uri)
   resource *r;
 
   if ( !uri[0] )			/* <> */
-    return new_resource(ts, ts->base_uri);
+    return new_resource(ts, ts->base_uri ? ts->base_uri : L"");
 
   if ( is_scheme_char(uri[0]) )		/* absolute uri */
   { for(s=&uri[1]; *s && is_scheme_char(*s); s++)
@@ -1254,7 +1273,7 @@ set_format(turtle_state *ts, format fmt)
   { switch(fmt)
     { case D_TRIG:
 	ts->default_graph = ts->current_graph;
-        ts->current_graph = NULL;
+	ts->current_graph = NULL;
 	/*FALLTHROUGH*/
       case D_TRIG_NO_GRAPH:
       case D_TURTLE:
@@ -1262,7 +1281,7 @@ set_format(turtle_state *ts, format fmt)
 	return TRUE;
       default:
 	assert(0);
-        return FALSE;
+	return FALSE;
     }
   }
 
@@ -1485,8 +1504,8 @@ put_graph(turtle_state *ts, term_t g)
     if ( (pos=ts->input->position) )
     { PL_put_variable(g);
       return PL_unify_term(g, PL_FUNCTOR, FUNCTOR_colon2,
-			        PL_ATOM, cg->v.r.handle,
-			        PL_INT,  (int)pos->lineno);
+				PL_ATOM, cg->v.r.handle,
+				PL_INT,  (int)pos->lineno);
     } else
     { return PL_put_atom(g, cg->v.r.handle);
     }
@@ -2368,8 +2387,8 @@ read_number(turtle_state *ts, string_buffer *num, number_type *numtype)
 	{ *numtype = NUM_DOUBLE;
 	  return read_exponent(ts, num);
 	}
-        default:
-        done:
+	default:
+	done:
 	  addBuf(num, EOS);
 	  return TRUE;
       }
@@ -2761,12 +2780,22 @@ starts_graph(turtle_state *ts)
 }
 
 
+#define	NO_GRAPH \
+	if ( graph_keyword_seen ) \
+	  return syntax_error(ts, "graph IRI expected after GRAPH keyword") \
+
 static int
-graph_or_final_predicate_object_list(turtle_state *ts, resource *r)
+graph_or_final_predicate_object_list(turtle_state *ts, resource *r,
+				     int *must_be_graph)
 { int rc;
+  int graph_keyword_seen = *must_be_graph;
+
+  *must_be_graph = FALSE;
 
   if ( !skip_ws(ts) )
+  { NO_GRAPH;
     return FALSE;
+  }
 
   if ( (rc = starts_graph(ts)) == TRUE )
   { switch ( ts->format )
@@ -2785,18 +2814,19 @@ graph_or_final_predicate_object_list(turtle_state *ts, resource *r)
       case D_TURTLE:
 	syntax_warning(ts, "Unexpected \"<graph> {\" in Turtle format "
 		       "(assuming TriG, ignoring graphs)");
-        set_format(ts, D_TRIG_NO_GRAPH);
+	set_format(ts, D_TRIG_NO_GRAPH);
 	/*FALLTHROUGH*/
       case D_TRIG_NO_GRAPH:
 	free_resource(ts, r);		/* discard graph name */
 	return next(ts) && statement(ts);
       default:
 	assert(0);
-        return FALSE;
+	return FALSE;
     }
   } else if ( rc == FALSE )
   { set_subject(ts, r, NULL);
 
+    NO_GRAPH;
     return final_predicate_object_list(ts);
   } else
     return FALSE;			/* error from starts_graph() */
@@ -2805,17 +2835,20 @@ graph_or_final_predicate_object_list(turtle_state *ts, resource *r)
 
 static int
 statement(turtle_state *ts)
-{ retry:
+{ int graph_keyword_seen = FALSE;
 
+retry:
   if ( !skip_ws(ts) )
     return FALSE;
 
   switch ( ts->current_char )
   { case '@':				/* directive */
-    { return next(ts) && directive(ts);
+    { NO_GRAPH;
+      return next(ts) && directive(ts);
     }
     case '}':
-    { if ( ts->format == D_TRIG )
+    { NO_GRAPH;
+      if ( ts->format == D_TRIG )
       { set_graph(ts, NULL, NULL);
 	return next(ts);			/* return } as empty triple set */
       } else if ( ts->format == D_TRIG_NO_GRAPH )
@@ -2824,11 +2857,12 @@ statement(turtle_state *ts)
 	return syntax_error(ts, "Unexpected \"}\" in Turtle format");
     }
     case '{':
-    { switch( ts->format )
+    { NO_GRAPH;
+      switch( ts->format )
       { case D_AUTO:
 	  set_format(ts, D_TRIG);
 	  /*FALLTHROUGH*/
-        case D_TRIG:
+	case D_TRIG:
 	  if ( !ts->current_graph )
 	  { if ( !next(ts) )
 	      return FALSE;
@@ -2844,7 +2878,7 @@ statement(turtle_state *ts)
 			 "(assuming TriG, ignoring graphs)");
 	  set_format(ts, D_TRIG_NO_GRAPH);
 	  /*FALLTHROUGH*/
-        case D_TRIG_NO_GRAPH:
+	case D_TRIG_NO_GRAPH:
 	  if ( !next(ts) )
 	    return FALSE;
 	  goto retry;
@@ -2854,12 +2888,13 @@ statement(turtle_state *ts)
     { resource *r;
 
       if ( (r=read_iri_ref(ts)) )
-	return graph_or_final_predicate_object_list(ts, r);
+	return graph_or_final_predicate_object_list(ts, r, &graph_keyword_seen);
       return FALSE;
     }
     case '[':
     { resource *r;
 
+      NO_GRAPH;
       if ( !next(ts) || !skip_ws(ts) )
 	return FALSE;
       if ( ts->current_char == ']' )
@@ -2879,6 +2914,7 @@ statement(turtle_state *ts)
     case '(':
     { resource *r;
 
+      NO_GRAPH;
       if ( (r=read_collection(ts)) )
       { set_subject(ts, r, NULL);
 	return final_predicate_object_list(ts);
@@ -2888,6 +2924,7 @@ statement(turtle_state *ts)
     case '_':
     { resource *r;
 
+      NO_GRAPH;
       if ( (r=read_blank_node_label(ts)) )
       { set_subject(ts, r, NULL);
 	return final_predicate_object_list(ts);
@@ -2907,19 +2944,21 @@ statement(turtle_state *ts)
 	  r = resolve_iri(ts, NULL, baseBuf(&pn_local));
 	  discardBuf(&pn_local);
 	  if ( r )
-	    return graph_or_final_predicate_object_list(ts, r);
+	    return graph_or_final_predicate_object_list(ts, r,
+							&graph_keyword_seen);
 	}
       } else
       { resource *r;
 
 	if ( (r=resolve_iri(ts, NULL, NULL)) )
-	  return graph_or_final_predicate_object_list(ts, r);
+	  return graph_or_final_predicate_object_list(ts, r, &graph_keyword_seen);
       }
 
       return FALSE;
     }
     case -1:
-    { return TRUE;				/* End of file */
+    { NO_GRAPH;
+      return TRUE;				/* End of file */
     }
     default:
     { string_buffer pn_prefix;
@@ -2927,12 +2966,31 @@ statement(turtle_state *ts)
       if ( read_pn_prefix(ts, &pn_prefix) )
       { if ( wcscasecmp(baseBuf(&pn_prefix), L"BASE") == 0 )
 	{ discardBuf(&pn_prefix);
+	  NO_GRAPH;
 	  return sparql_base_directive(ts);
 	}
 	if ( wcscasecmp(baseBuf(&pn_prefix), L"PREFIX") == 0 )
 	{ discardBuf(&pn_prefix);
+	  NO_GRAPH;
 	  return sparql_prefix_directive(ts);
 	}
+	if ( wcscasecmp(baseBuf(&pn_prefix), L"GRAPH") == 0 &&
+	     ts->current_char != ':' )
+	{ discardBuf(&pn_prefix);
+	  NO_GRAPH;
+	  if ( ts->format == D_TURTLE )
+	  { syntax_warning(ts, "Unexpected \"GRAPH\" in Turtle format "
+			 "(assuming TriG, ignoring graphs)");
+	    set_format(ts, D_TRIG_NO_GRAPH);
+	  } else if ( ts->format == D_AUTO )
+	  { set_format(ts, D_TRIG);
+	  }
+	  if ( !next(ts) )
+	    return FALSE;
+	  graph_keyword_seen = TRUE;
+
+	  goto retry;
+	}
 
 	if ( ts->current_char == ':' )
 	{ if ( !next(ts) )
@@ -2950,7 +3008,8 @@ statement(turtle_state *ts)
 	      discardBuf(&pn_local);
 	      if ( r )
 	      { discardBuf(&pn_prefix);
-		return graph_or_final_predicate_object_list(ts, r);
+		return graph_or_final_predicate_object_list(ts, r,
+							    &graph_keyword_seen);
 	      }
 	    }
 	  } else
@@ -2960,10 +3019,12 @@ statement(turtle_state *ts)
 	    discardBuf(&pn_prefix);
 
 	    if ( r )
-	      return graph_or_final_predicate_object_list(ts, r);
+	      return graph_or_final_predicate_object_list(ts, r,
+							  &graph_keyword_seen);
 	  }
 	} else
 	{ discardBuf(&pn_prefix);
+	  NO_GRAPH;
 	  return syntax_error(ts, "Expected \":\"");
 	}
       }
@@ -3397,7 +3458,7 @@ turtle_format(term_t parser, term_t format)
       case D_TRIG_NO_GRAPH: a = ATOM_trig;   break;
       default:
 	assert(0);
-        return FALSE;
+	return FALSE;
     }
 
     return PL_unify_atom(format, a);
