@@ -68,7 +68,9 @@
             sweep_predicate_completion_candidates/2,
             sweep_exportable_predicates/2,
             sweep_interrupt/0,
-            sweep_string_to_atom/2
+            sweep_string_to_atom/2,
+            sweep_file_path_in_library/2,
+            sweep_file_missing_dependencies/2
           ]).
 
 :- use_module(library(pldoc)).
@@ -84,6 +86,8 @@
 :- use_module(library(lynx/html_text)).
 :- use_module(library(http/html_write)).
 :- use_module(library(prolog_pack)).
+:- use_module(library(prolog_deps)).
+:- use_module(library(dcg/high_order)).
 
 :- if(exists_source(library(help))).
 :- use_module(library(help)).
@@ -349,23 +353,15 @@ sweep_predicate_apropos(Query0, Matches) :-
             Tail).
 
 sweep_predicate_location_(H, Path, Line) :-
-    predicate_property(H, file(Path0)),
-    predicate_property(H, line_count(Line)),
+    xref_defined(Path0, H, How),
+    xref_definition_line(How, Line),
     !,
     atom_string(Path0, Path).
 sweep_predicate_location_(H, Path, Line) :-
-    (   xref_defined(Path0, H, How),
-        xref_definition_line(How, Line)
-    ->  true
-    ;   xref_defined(Path0, H, _), Line = []
-    ),
+    predicate_property(H, file(Path0)),
+    predicate_property(H, line_count(Line)),
     atom_string(Path0, Path).
 
-sweep_predicate_location_(M, H, Path, Line) :-
-    predicate_property(M:H, file(Path0)),
-    predicate_property(M:H, line_count(Line)),
-    !,
-    atom_string(Path0, Path).
 sweep_predicate_location_(M, H, Path, Line) :-
     (   xref_defined(Path0, M:H, How),
         xref_definition_line(How, Line)
@@ -373,9 +369,12 @@ sweep_predicate_location_(M, H, Path, Line) :-
     ;   xref_defined(Path0, H, How),
         xref_definition_line(How, Line),
         xref_module(Path0, M)
-    ->  true
-    ;   xref_defined(Path0, M:H, _), Line = []
     ),
+    !,
+    atom_string(Path0, Path).
+sweep_predicate_location_(M, H, Path, Line) :-
+    predicate_property(M:H, file(Path0)),
+    predicate_property(M:H, line_count(Line)),
     atom_string(Path0, Path).
 
 
@@ -466,7 +465,7 @@ sweep_color_normalized_(_, Goal0, [Kind0,Head|_], [Goal,Kind,F,N]) :-
     sweep_color_goal(Goal0),
     !,
     atom_string(Goal0, Goal),
-    term_string(Kind0, Kind),
+    sweeprolog_goal_kind_normalized(Kind0, Kind),
     (   (   var(Head)
         ->  true
         ;   Head == []
@@ -504,6 +503,13 @@ sweep_color_normalized_(_, file_no_depend, [File0|_], ["file_no_depend"|File]) :
 sweep_color_normalized_(_, Nom0, _, Nom) :-
     atom_string(Nom0, Nom).
 
+sweeprolog_goal_kind_normalized(autoload(Path0), ["autoload"|Path]) :-
+    !,
+    absolute_file_name(Path0, Path1, [extensions([pl])]),
+    atom_string(Path1, Path).
+sweeprolog_goal_kind_normalized(Kind0, Kind) :-
+    term_string(Kind0, Kind).
+
 sweep_color_goal(goal).
 sweep_color_goal(goal_term).
 sweep_color_goal(head).
@@ -710,8 +716,9 @@ strip_det(Mode is _, Mode) :- !.
 strip_det(//(Mode), Mode) :- !.
 strip_det(Mode, Mode).
 
-sweep_functor_arity_pi([F0,A], PI)   :-
-    !,
+sweep_functor_arity_pi([M,F0,A|_], PI) :-
+    !, atom_string(F, F0), term_string(M:F/A, PI).
+sweep_functor_arity_pi([F0,A|Path0], PI) :-
     atom_string(F, F0),
     pi_head(F/A, Head),
     sweep_current_module(M0),
@@ -720,11 +727,13 @@ sweep_functor_arity_pi([F0,A], PI)   :-
     ->  T = M:F/A
     ;   xref_defined(_, Head, imported(Other)), xref_module(Other, M)
     ->  T = M:F/A
+    ;   string(Path0),
+        atom_string(Path, Path0),
+        xref_defined(Path, Head, _)
+    ->  T = M0:F/A
     ;   T = F/A
     ),
     term_string(T, PI).
-sweep_functor_arity_pi([M,F0,A], PI) :-
-    atom_string(F, F0), term_string(M:F/A, PI).
 
 sweep_current_module(Module) :-
     sweep_main_thread,
@@ -764,23 +773,39 @@ sweep_atom_collection(Sub, Col) :-
             ),
             Col).
 
-sweep_predicate_completion_candidates(_, Ps) :-
+sweep_predicate_completion_candidates(D, Ps) :-
+    integer(D),
+    sweep_current_module(M),
     findall(H,
-            (   sweep_current_module(M),
-                @(predicate_property(H, visible), M)
+            (   (   @(predicate_property(H, visible), M)
+                ;   xref_defined(_, H, _)
+                ),
+                pi_head(_/N, H),
+                N - D >= 0
             ),
             Hs),
-    maplist(sweep_format_predicate, Hs, Ps).
+    maplist(sweep_format_predicate(M, D), Hs, Ps).
 
-sweep_format_predicate(H, [S|SP]) :-
-    term_variables(H, Vs),
-    maplist(=('$VAR'('_')), Vs),
+sweep_format_predicate(M0, D, H0, [S|SP]) :-
+    pi_head(F/N0, H0),
+    N is N0 - D,
+    length(NamedArgs, N),
+    append(NamedArgs, _, OpenNamedArgs),
+    (   @(predicate_property(H0, implementation_module(M)), M0),
+        predicate_argument_names(M:F/N0, As)
+    ->  maplist(name_variable, As, Vs), OpenNamedArgs = Vs
+    ;   maplist(=('$VAR'('_')), NamedArgs)
+    ),
+    !,
+    H =.. [F|NamedArgs],
     term_string(H, S, [quoted(true),
                        character_escapes(true),
                        spacing(next_argument),
                        numbervars(true)]),
     term_string(_, S, [subterm_positions(SP)]).
 
+name_variable(N, V) :- V = '$VAR'(N).
+
 sweep_context_callable([H|T], R) :-
     H = [F0|_],
     atom_string(F, F0),
@@ -789,43 +814,55 @@ sweep_context_callable([H|T], R) :-
     ;   current_op(1200, _, F)
     ),
     !,
-    sweep_context_callable_(T, R).
+    (   F == (-->)
+    ->  R0 = 2
+    ;   R0 = 0
+    ),
+    sweep_context_callable_(T, R0, 0, R).
 sweep_context_callable([_|T], R) :-
     sweep_context_callable(T, R).
 
-sweep_context_callable_([], true) :- !.
-sweep_context_callable_([[":"|2]], true) :- !.
-sweep_context_callable_([["("|_]|T], R) :-
-    sweep_context_callable_(T, R).
-sweep_context_callable_([H|T], R) :-
+sweep_context_callable_([], R0, R1, R) :- R is R0 + R1, !.
+sweep_context_callable_([[":"|2]], R0, R1, R) :- R is R0 + R1, !.
+sweep_context_callable_([["("|_]|T], R0, R1, R) :-
+    !,
+    sweep_context_callable_(T, R0, R1, R).
+sweep_context_callable_([["{"|_]|T], 2, R1, R) :-
+    !,
+    sweep_context_callable_(T, 0, R1, R).
+sweep_context_callable_([H|T], R0, _, R) :-
     H = [F0|N],
     atom_string(F, F0),
-    (   sweep_context_callable_arg(F, N)
-    ->  sweep_context_callable_(T, R)
-    ;   R = []
-    ).
+    sweep_context_callable_arg(F, N, R1),
+    sweep_context_callable_(T, R0, R1, R).
 
-sweep_context_callable_arg(Neck, _) :-
+sweep_context_callable_arg((-->), _, 2) :- !.
+sweep_context_callable_arg(Neck, _, 0) :-
     (   xref_op(_, op(1200, _, Neck))
     ->  true
     ;   current_op(1200, _, Neck)
-    ).
-sweep_context_callable_arg(F, N) :-
-    (   current_predicate(F/M), pi_head(F/M,Head)
-    ;   xref_defined(_, Head, _), pi_head(F/M,Head)
     ),
+    !.
+sweep_context_callable_arg(F, N, R) :-
+    sweep_current_module(Mod),
+    (   @(predicate_property(Head, visible), Mod)
+    ;   xref_defined(_, Head, _)
+    ),
+    pi_head(F/M,Head),
     M >= N,
-    catch(infer_meta_predicate(Head, Spec),
-          error(permission_error(access, private_procedure, _),
-                context(system:clause/2, _)),
-          false),
+    (   @(predicate_property(Head, meta_predicate(Spec)), Mod)
+    ;   catch(infer_meta_predicate(Head, Spec),
+              error(permission_error(access, private_procedure, _),
+                    context(system:clause/2, _)),
+              false)
+    ),
     arg(N, Spec, A),
-    callable_arg(A).
+    callable_arg(A, R).
 
-callable_arg(N) :- integer(N), !.
-callable_arg(^) :- !.
-callable_arg(//) :- !.
-callable_arg(:) :- !.
+callable_arg(N,    N) :- integer(N), !.
+callable_arg((^),  0) :- !.
+callable_arg((//), 2) :- !.
+callable_arg((:),  0) :- !.
 
 sweep_exportable_predicates(Path0, Preds) :-
     atom_string(Path, Path0),
@@ -848,3 +885,55 @@ sweep_string_to_atom(String, AtomString) :-
     format(string(AtomString),
            "~W",
            [Atom, [quoted(true), character_escapes(true)]]).
+
+sweep_file_path_in_library(Path, Spec) :-
+    file_name_on_path(Path, Spec0),
+    prolog_deps:segments(Spec0, Spec1),
+    term_string(Spec1, Spec).
+
+predicate_argument_names(M:F/A, Args) :-
+    doc_comment(M:F/A, _, _, C),
+    comment_modes(C, ModeAndDets),
+    member(ModeAndDet, ModeAndDets),
+    strip_det(ModeAndDet, Head),
+    Head =.. [_|Args0],
+    length(Args0, A),
+    maplist(strip_mode_and_type, Args0, Args).
+
+strip_mode_and_type(S, N), compound(S) => arg(1, S, N0), strip_type(N0, N).
+strip_mode_and_type(S, N) => strip_type(S, N).
+
+strip_type(N:_, N) :- !.
+strip_type(N, N).
+
+sweep_file_missing_dependencies(File0, Deps) :-
+    atom_string(File, File0),
+    file_autoload_directives(File, Directives, [missing(true)]),
+    phrase(dep_directives(Directives), Deps).
+
+dep_directives(Directives) --> sequence(dep_directive, Directives).
+
+dep_directive(:- use_module(Spec)) -->
+    !,
+    {      absolute_file_name(Spec, Path0,
+                           [ file_type(prolog),
+                             access(read)
+                           ]),
+           atom_string(Path0, Path)
+    },
+    [[Path, [], "use_module"]].
+dep_directive(:- Directive) -->
+    {   compound_name_arguments(Directive, Kind0, [Spec, ImportList]),
+        atom_string(Kind0, Kind),
+        absolute_file_name(Spec, Path0,
+                           [ file_type(prolog),
+                             access(read)
+                           ]),
+        atom_string(Path0, Path)
+    },
+    sequence(dep_import(Path, Kind), ImportList).
+
+dep_import(Path, Kind, PI0) -->
+    {   term_string(PI0, PI)
+    },
+    [[Path, PI, Kind]].
