@@ -5,6 +5,10 @@
 (remove-hook 'flymake-diagnostic-functions
              #'flymake-proc-legacy-flymake)
 
+(add-hook 'sweeprolog-mode-hook (lambda ()
+                                  (setq-local indent-tabs-mode nil
+                                              inhibit-message t)))
+
 (defconst sweeprolog-tests-greeting
   "Hello from Elisp from Prolog from Elisp from Prolog from Elisp!")
 
@@ -17,7 +21,7 @@
     (cdr sol)))
 
 (defun sweeprolog-tests-greet-1 ()
-  (message sweeprolog-tests-greeting))
+  sweeprolog-tests-greeting)
 
 (ert-deftest elisp->prolog->elisp->prolog->elisp ()
   "Tests calling Elisp from Prolog from Elisp from Prolog from Elisp."
@@ -229,6 +233,71 @@ baz(Baz) :- Baz = opaque
 "
                      ))))
 
+(ert-deftest complete-non-terminal ()
+  "Tests completing DCG non-terminals."
+  (let ((temp (make-temp-file "sweeprolog-test"
+                              nil
+                              ".pl"
+                              "
+barbaz --> foo.
+
+foo --> barb"
+                              )))
+    (find-file-literally temp)
+    (sweeprolog-mode)
+    (goto-char (point-max))
+    (call-interactively #'completion-at-point)
+    (should (string= (buffer-string)
+                     "
+barbaz --> foo.
+
+foo --> barbaz"
+
+                     ))
+    (insert ".\n\nfoo => barb")
+    (call-interactively #'completion-at-point)
+    (should (string= (buffer-string)
+                     "
+barbaz --> foo.
+
+foo --> barbaz.
+
+foo => barbaz(_, _)"
+
+                     ))))
+
+(ert-deftest complete-predicate-with-args ()
+  "Tests completing predicate calls."
+  (let ((temp (make-temp-file "sweeprolog-test"
+                              nil
+                              ".pl"
+                              "
+:- module(foobarbaz, []).
+
+%!  foobarbaz(:Bar, ?Baz:integer) is det.
+
+foobarbaz(_, 5) :- spam.
+
+spam :- foobarb
+"
+                              )))
+    (find-file-literally temp)
+    (sweeprolog-mode)
+    (goto-char (point-max))
+    (backward-char)
+    (call-interactively #'completion-at-point)
+    (should (string= (buffer-string)
+                     "
+:- module(foobarbaz, []).
+
+%!  foobarbaz(:Bar, ?Baz:integer) is det.
+
+foobarbaz(_, 5) :- spam.
+
+spam :- foobarbaz(Bar, Baz)
+"
+                     ))))
+
 (ert-deftest complete-predicate ()
   "Tests completing predicate calls."
   (let ((temp (make-temp-file "sweeprolog-test"
@@ -522,7 +591,7 @@ foo.
 foo :- Body.
 "))))
 
-(ert-deftest dwim-next-clause-module-qualified-cdg ()
+(ert-deftest dwim-next-clause-module-qualified-dcg ()
   "Tests inserting new module-qualified DCG non-terminal."
   (let ((temp (make-temp-file "sweeprolog-test"
                               nil
@@ -626,8 +695,162 @@ foo :- bar.
 foo :- Body.
 "))))
 
-(ert-deftest dwim-define-nested-phrase- ()
-  "Tests complex undefined predicate scenario"
+(ert-deftest update-dependencies-no-autoload ()
+  "Tests making adding a use_module/1 directive."
+  (let ((temp (make-temp-file "sweeprolog-test"
+                              nil
+                              "pl"
+                              "
+:- module(foo, [bar/1]).
+
+/** <module> Foo
+
+*/
+
+bar(X) :- arithmetic_function(X).
+"
+                              )))
+    (find-file-literally temp)
+    (sweeprolog-mode)
+    (call-interactively #'sweeprolog-update-dependencies)
+    (should (string= (buffer-string)
+                              "
+:- module(foo, [bar/1]).
+
+/** <module> Foo
+
+*/
+
+:- use_module(library(arithmetic), [ arithmetic_function/1
+                                   ]).
+
+bar(X) :- arithmetic_function(X).
+"))))
+
+(ert-deftest append-dependencies ()
+  "Tests making implicit autoloads explicit with existing directive."
+  (let ((temp (make-temp-file "sweeprolog-test"
+                              nil
+                              "pl"
+                              "
+:- module(foo, [bar/1]).
+
+/** <module> Foo
+
+*/
+
+:- use_module(library(lists), [ member/2
+                              ]).
+
+bar(X) :- member(X, [1,2,3]).
+bar(X) :- permutation(X, [1,2,3]).
+"
+                              )))
+    (find-file-literally temp)
+    (sweeprolog-mode)
+    (call-interactively #'sweeprolog-update-dependencies)
+    (should (string= (buffer-string)
+                     "
+:- module(foo, [bar/1]).
+
+/** <module> Foo
+
+*/
+
+:- use_module(library(lists), [ member/2,
+                                permutation/2
+                              ]).
+
+bar(X) :- member(X, [1,2,3]).
+bar(X) :- permutation(X, [1,2,3]).
+"
+                     ))))
+
+(ert-deftest update-dependencies-autoload-from-package ()
+  "Tests making implicit autoloads from a package explicit."
+  (let ((temp (make-temp-file "sweeprolog-test"
+                              nil
+                              "pl"
+                              "
+:- module(foo, [bar/1]).
+
+/** <module> Foo
+
+*/
+
+bar(X) :- http_open(X, X, X).
+"
+                              )))
+    (find-file-literally temp)
+    (sweeprolog-mode)
+    (call-interactively #'sweeprolog-update-dependencies)
+    (should (string= (buffer-string)
+                     "
+:- module(foo, [bar/1]).
+
+/** <module> Foo
+
+*/
+
+:- autoload(library(http/http_open), [ http_open/3
+                                     ]).
+
+bar(X) :- http_open(X, X, X).
+"))))
+
+(ert-deftest update-dependencies ()
+  "Tests making implicit autoloads explicit."
+  (let ((temp (make-temp-file "sweeprolog-test"
+                              nil
+                              "pl"
+                              "
+:- module(foo, [bar/1]).
+
+/** <module> Foo
+
+*/
+
+bar(X) :- member(X, [1,2,3]).
+"
+                              )))
+    (find-file-literally temp)
+    (sweeprolog-mode)
+    (call-interactively #'sweeprolog-update-dependencies)
+    (should (string= (buffer-string)
+                     "
+:- module(foo, [bar/1]).
+
+/** <module> Foo
+
+*/
+
+:- autoload(library(lists), [ member/2
+                            ]).
+
+bar(X) :- member(X, [1,2,3]).
+"
+
+                     ))
+    (goto-char (point-max))
+    (insert "bar(X) :- permutation(X, [1,2,3]).")
+    (call-interactively #'sweeprolog-update-dependencies)
+    (should (string= (buffer-string)
+                     "
+:- module(foo, [bar/1]).
+
+/** <module> Foo
+
+*/
+
+:- autoload(library(lists), [ member/2,
+                              permutation/2
+                            ]).
+
+bar(X) :- member(X, [1,2,3]).
+bar(X) :- permutation(X, [1,2,3])."))))
+
+(ert-deftest dwim-define-nested-phrase ()
+  "Tests complex undefined predicate scenario."
   (let ((temp (make-temp-file "sweeprolog-test"
                               nil
                               "pl"
@@ -801,6 +1024,53 @@ scasp_and_show(Q, Model, Tree) :-
     (sweeprolog-end-of-top-term)
     (should (= (point) 252))))
 
+(ert-deftest beginning-of-predicate-definition-near-bob ()
+  "Test finding the beginning of the first predicate definition."
+  (with-temp-buffer
+    (sweeprolog-mode)
+    (insert "foo :- bar.")
+    (goto-char (point-min))
+    (sweeprolog-beginning-of-predicate-at-point)
+    (should (= (point) (point-min)))))
+
+(ert-deftest align-spacs-in-line-comment ()
+  "Test using `sweeprolog-align-spaces' in a line comment."
+  (with-temp-buffer
+    (sweeprolog-mode)
+    (insert "
+%!  foo is det.
+%
+%")
+    (sweeprolog-align-spaces)
+    (should (string= (buffer-string)
+                     "
+%!  foo is det.
+%
+%   "))))
+
+(ert-deftest auto-fill-pldoc-comments ()
+  "Test writing PlDoc comments with `auto-fill-mode' enable."
+  (with-temp-buffer
+    (sweeprolog-mode)
+    (auto-fill-mode)
+    (seq-do (lambda (c)
+              (let ((last-command-event c))
+                (call-interactively #'self-insert-command)))
+            "
+%!  foobar is det.
+%
+%   Nam vestibulum accumsan nisl.  Donec pretium posuere tellus.  Aenean in sem ac leo mollis blandit.  Nam a sapien.  Proin quam nisl, tincidunt et, mattis eget, convallis nec, purus.
+"
+            )
+    (should (string= (buffer-string)
+                     "
+%!  foobar is det.
+%
+%   Nam vestibulum accumsan nisl.  Donec pretium posuere tellus.
+%   Aenean in sem ac leo mollis blandit.  Nam a sapien.  Proin quam
+%   nisl, tincidunt et, mattis eget, convallis nec, purus.
+"))))
+
 (ert-deftest electric-layout ()
   "Test `sweeprolog-electric-layout-mode'."
   (with-temp-buffer
@@ -864,12 +1134,48 @@ test_bindings(Name-Value) -->
 "
                      ))))
 
+(ert-deftest infer-indent-style ()
+  "Test inferring indentation style from buffer contents."
+  (with-temp-buffer
+    (sweeprolog-mode)
+    (insert "
+foo :-
+  bar.")
+    (sweeprolog-infer-indent-style)
+    (should (= sweeprolog-indent-offset 2))
+    (should (not indent-tabs-mode)))
+  (with-temp-buffer
+    (sweeprolog-mode)
+    (insert "
+foo :-
+\tbar.")
+    (sweeprolog-infer-indent-style)
+    (should (= sweeprolog-indent-offset tab-width))
+    (should indent-tabs-mode)))
+
+(ert-deftest custom-indentation ()
+  "Test forcefully setting custom indentation levels."
+  (with-temp-buffer
+    (sweeprolog-mode)
+    (insert "
+foo :-
+    repeat,
+      bar,
+      baz")
+    (call-interactively #'indent-for-tab-command)
+    (should (string= (buffer-substring-no-properties (point-min) (point-max))
+                     "
+foo :-
+    repeat,
+      bar,
+      baz"))))
 
 (defun sweeprolog-test-indentation (given expected)
   (with-temp-buffer
     (sweeprolog-mode)
     (insert given)
-    (indent-region-line-by-line (point-min) (point-max))
+    (let ((inhibit-message t))
+      (indent-region-line-by-line (point-min) (point-max)))
     (should (string= (buffer-substring-no-properties (point-min) (point-max))
                      expected))))
 
@@ -877,24 +1183,25 @@ test_bindings(Name-Value) -->
   (with-temp-buffer
     (sweeprolog-mode)
     (insert given)
-    (let ((callable (sweeprolog-context-callable-p)))
-      (should (if expected
-                  callable
-                (not callable))))))
+    (should (equal expected (sweeprolog-context-callable-p)))))
 
 (ert-deftest context-callable ()
   "Test recognizing callable contexts."
+  (sweeprolog-test-context-callable-p "foo(Bar) :- include( " 1)
+  (sweeprolog-test-context-callable-p "foo(Bar) --> " 2)
+  (sweeprolog-test-context-callable-p "foo(Bar) --> {include(" 1)
+  (sweeprolog-test-context-callable-p "foo(Bar) --> {include(phrase(" 2)
   (sweeprolog-test-context-callable-p "foo" nil)
   (sweeprolog-test-context-callable-p "foo(" nil)
   (sweeprolog-test-context-callable-p "foo(bar)" nil)
-  (sweeprolog-test-context-callable-p "foo(bar) :- " t)
+  (sweeprolog-test-context-callable-p "foo(bar) :- " 0)
   (sweeprolog-test-context-callable-p "foo(bar) :- baz(" nil)
   (sweeprolog-test-context-callable-p "foo(bar) :- baz(bar" nil)
-  (sweeprolog-test-context-callable-p "foo(bar) :- baz(bar), " t)
+  (sweeprolog-test-context-callable-p "foo(bar) :- baz(bar), " 0)
   (sweeprolog-test-context-callable-p "foo(bar) :- baz(bar), findall(" nil)
   (sweeprolog-test-context-callable-p "foo(bar) :- baz(bar), findall(X" nil)
-  (sweeprolog-test-context-callable-p "foo(bar) :- baz(bar), findall(X," t)
-  (sweeprolog-test-context-callable-p "foo(bar) :- baz(bar), findall(X, false" t)
+  (sweeprolog-test-context-callable-p "foo(bar) :- baz(bar), findall(X," 0)
+  (sweeprolog-test-context-callable-p "foo(bar) :- baz(bar), findall(X, false" 0)
   (sweeprolog-test-context-callable-p "foo(bar) :- baz(bar), findall(X, false," nil)
   (sweeprolog-test-context-callable-p "foo(bar) :- baz(bar), findall(X, false, Xs). " nil))
 
@@ -902,6 +1209,15 @@ test_bindings(Name-Value) -->
   "Tests indentation rules."
   (sweeprolog-test-indentation
    "
+colourise_declaration(Module:Goal, table, TB,
+                      term_position(_,_,QF,QT,
+"
+   "
+colourise_declaration(Module:Goal, table, TB,
+                      term_position(_,_,QF,QT,
+")
+  (sweeprolog-test-indentation
+   "
 some_functor(
 arg1,
 arg2,
@@ -1133,6 +1449,28 @@ foo"
    "
 functor(
     foo")
-  )
+  (sweeprolog-test-indentation
+   "
+replace_key_value(Replacement, Key - AtVar, Out, Used0, Used1),
+atom_concat(@, Var, AtVar) =>
+foo.
+"
+   "
+replace_key_value(Replacement, Key - AtVar, Out, Used0, Used1),
+  atom_concat(@, Var, AtVar) =>
+    foo.
+"
+   )
+  (sweeprolog-test-indentation
+   "
+head,
+right_hand_context -->
+body.
+"
+   "
+head,
+  right_hand_context -->
+    body.
+"))
 
 ;;; sweeprolog-tests.el ends here
